#ifndef FERI_PDE3
#define FERI_PDE3
// --------------------------------------------
// -- Rigorous Numerics for Dissipative PDEs --
// --------------------------------------------
// --------------- Main Program ---------------
// --------------------------------------------
// by Ferenc Bartha
// http://www.math.u-szeged.hu/~barfer
// --------------------------------------------

#include "main.hpp"
#include "tools.hpp"
#include "structures.hpp"
#include "initbounds.hpp"
#include "vectorfield.hpp"
#include "typedefs.hpp"
#include "fixedpoint.hpp"
#include "nonrigorous.hpp"

// --------------------------------------------

// Resetting screen
void resetGraphics(const double &ScreenSize)
{
	rootFrame.Clear();
	rootFrame.line(-ScreenSize,0.,ScreenSize,0.);
	rootFrame.line(0.,-ScreenSize,0.,ScreenSize);
}

// --------------------------------------------

int main(int argc, char* argv[])
{
	MainType TheMain;
	TailType TheTail;
	BoundsType bounds;
	
	interval alpha, beta, gamma, nu, L;
	double Delta;

	// reading the command line parameters
	if (argc > 1) {

		initFromFile(alpha, beta, gamma, nu, L, TheMain, TheTail, argv[2]);

		// initialize bounds
		bounds.W = VectorType(TheMain);
		bounds.T = TheTail;

		// output file
		openOutput(argv[3]);

		if (std::strcmp(argv[1], "proof") == 0 ) {
			Delta = atof(argv[4]);
		
			Indent_AddToOutput();
			ostream << "The paramteres are set to:" << std::endl << "alpha = " << alpha << std::endl << "beta = " << beta << std::endl << "gamma = " << gamma << std::endl << "nu = " << nu << std::endl << "L = " << L << std::endl << "DELTA = " << Delta << std::endl;

		} else {
		
			Indent_AddToOutput();
			ostream << "The paramteres are set to:" << std::endl << "alpha = " << alpha << std::endl << "beta = " << beta << std::endl << "gamma = " << gamma << std::endl << "nu = " << nu << std::endl << "L = " << L << std::endl;

		}

	}


	// creating the vectorfield
	inclusion_Term vField(alpha, beta, gamma, nu, L);

// --------------------------------------------

	// entering graphics mode
	#ifdef FERI_GRAPHICS
		openGraphics(700,700,WHITE,BLACK);
		double SCREENSIZE = 40;
		rootFrame.setWorldCoord(-SCREENSIZE,-SCREENSIZE,SCREENSIZE,SCREENSIZE);

		resetGraphics(SCREENSIZE);
	#endif

// --------------------------------------------

	// trying to prove existence of a fixed point
	if (std::strcmp(argv[1], "proof") == 0 ) {
		bool result;
		FixedPointProof proof;

		result = proof.Validate_FixedPoint(bounds, TheTail, vField, Delta);

		#ifdef FERI_GRAPHICS
			if (result) {

				rootFrame.Clear();
				rootFrame << "Fixed point is verified.";
				rootFrame.line(-40.0,0.,40.0,0.);
				rootFrame.line(0.,-40.0,0.,40.0);
				vField.sin_plot(bounds, 100);
				waitBt();

			} else {

				rootFrame << "\nVerification failed.";
				waitBt();

			}

			waitBt();
		#endif
	}

// --------------------------------------------

	// finding a non rigorous candidate
	if (not(std::strcmp(argv[1], "proof") == 0 )) {

		// reset the rounding
		capd :: rounding :: DoubleRounding :: roundNearest();

		// initial guess
		nRVector xinit;
		for (int i = 1; i <= nR; ++i )
			xinit(i) = bounds(i).leftBound();

		// the non-rigorous projection
		nRVectorField my_nRField(alpha.leftBound(), beta.leftBound(), gamma.leftBound(), nu.leftBound(), L.leftBound());
		
		// reset the rounding
		capd :: rounding :: DoubleRounding :: roundNearest();

		// using Newtons method
		if (std::strcmp(argv[1], "newton") == 0 ) {

			Indent_AddToOutput();
			ostream << "Finding a non-rigorous candidate solution using Newtons method, dimension is: " << nR << std::endl;

			nonRig_simulate(my_nRField, true, xinit);

		// using a Taylor integrator		
		} else if (std::strcmp(argv[1], "integrate") == 0 ) {

			Indent_AddToOutput();
			ostream << "Finding a non-rigorous candidate solution using integration, dimension is: " << nR << std::endl;

			nonRig_simulate(my_nRField, false, xinit);

		}
	}

// --------------------------------------------

	// closing graphics mode
	#ifdef FERI_GRAPHICS
		closeGraphics();
		return 0;
	#endif
}

// ~ end of file ~
#endif
