#ifndef FERI_FIXP
#define FERI_FIXP
// --------------------------------------------
// -- Rigorous Numerics for Dissipative PDEs --
// --------------------------------------------
// ---------- Proof for Fixed Point -----------
// --------------------------------------------
// by Ferenc Bartha
// http://www.math.u-szeged.hu/~barfer
// --------------------------------------------

#include "main.hpp"
#include "tools.hpp"
#include "structures.hpp"
#include "vectorfield.hpp"
#include "typedefs.hpp"

// --------------------------------------------
// Proof for a Fixed Point
class FixedPointProof
{

public:

// --------------------------------------------
// the approximate jacobian at the approximate fixed point
void Approximate_Jacobian(const BoundsType &bounds, const inclusion_Term &vField, dMatrix &dDPmFx0)
{
	Indent_PlusLength();

	MatrixType DPmFx0;



	DPmFx0 = midMatrix(vField[bounds.W]);

	for (int i = 1; i <= m; ++i)
		for (int j = 1; j <= m; ++j)
			dDPmFx0(i,j) = DPmFx0(i,j).rightBound();

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "The approximate Jacobian is:" << std::endl << dDPmFx0 << std::endl;
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// eigenvalue decomposition
void Eigenvalue_Decomposition(const dMatrix &dDPmFx0, dMatrix &dvectReal, dMatrix &dvectIm, dVector &deigReal, dVector &deigIm)
{
	Indent_PlusLength();

	dVector t_column;


	// reset the rounding
	capd :: rounding :: DoubleRounding :: roundNearest();

	// approximate eigenvectors and eigenvalues
	Indent_AddToOutput();
	ostream << "Approximate eigendecomposition of the Jacobian...";
	capd::alglib::computeEigenvaluesAndEigenvectors( dDPmFx0, deigReal, deigIm, dvectReal, dvectIm );
	ostream << "done" << std::endl;



	// sorting according to the real part of the eigenvalues
	Indent_AddToOutput();
	ostream << "Sorting according to the real part of the eigenvalues...";
	for (int i = 1; i <= m - 1; ++i)
		for (int j = m; j > i; --j)
			if ( deigReal(i) < deigReal(j) ) {
				std::swap( deigReal(i), deigReal(j) );
				std::swap( deigIm(i), deigIm(j) );

				// columns are indexed from 0 to num - 1
				t_column = dvectReal.column(i - 1);
				dvectReal.column(i - 1) = dvectReal.column(j - 1);
				dvectReal.column(j - 1) = t_column;

				t_column = dvectIm.column(i - 1);
				dvectIm.column(i - 1) = dvectIm.column(j - 1);
				dvectIm.column(j - 1) = t_column;
			}
	ostream << "done" << std::endl;

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "The real part of the approximate eigenvalues:" << std::endl << deigReal << std::endl;
		Indent_AddToOutput();
		ostream << "The imaginary part of the approximate eigenvalues:" << std::endl << deigIm << std::endl;
		Indent_AddToOutput();
		ostream << "The real part of the approximate eigenvectors:" << std::endl << dvectReal << std::endl;
		Indent_AddToOutput();
		ostream << "The imaginary part of the approximate eigenvectors:" << std::endl << dvectIm << std::endl;
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// taking complex directions into account
void Complex_Directions(bVector &is_cpair_next, dMatrix &dvectReal, const dMatrix &dvectIm, const dVector &deigReal, const dVector &deigIm)
{
	Indent_PlusLength();

	bool previous_was_complex;
	bool there_is_complex;


	// marking complex directions
	previous_was_complex = false;
	there_is_complex = false;
	for (int i = 1; i < m; ++i) {
		if ( deigIm(i) != 0 ) {
				if ( !previous_was_complex ) {
					is_cpair_next(i) = true;
					there_is_complex = true;
				}
				previous_was_complex = not(previous_was_complex);
		} else {
			is_cpair_next(i) = false;
			previous_was_complex = false;
		}
	}
	is_cpair_next(m) = false;

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "First part of complex directions: " << std::endl << is_cpair_next << std::endl;
	#endif

	// substituting columns from the imaginary part if necessary
	if (there_is_complex) {
		Indent_AddToOutput();
		ostream << "Substituting columns from the imaginary part...";

		for (int i = 1; i < m; ++i)
			if ( is_cpair_next(i) )
				dvectReal.column(i) = dvectIm.column(i - 1);			// columns are indexed from 0 to num - 1
		ostream << "done" << std::endl;
	}

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "The new foundamental matrix is:" << std::endl << dvectReal << std::endl;
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// defining the new base
//  x = old, y = new coordinates
//  y = newBase * (x - newBaseOrigin), x = newBaseInv * y + newBaseOrigin
void Define_Base(const BoundsType &bounds, dMatrix &dvectReal, MatrixType &newBase, MatrixType &newBaseInv, VectorType &newBaseOrigin)
{
	dMatrix dnewBase;
	
	// reset the rounding
	capd :: rounding :: DoubleRounding :: roundNearest();

	newBaseInv = dvectReal;
	dnewBase = capd::matrixAlgorithms::gaussInverseMatrix(dvectReal);
	//dnewBase = capd::matrixAlgorithms::krawczykInverse(dvectReal);
	newBase = MatrixType(dnewBase);
	newBaseOrigin = midVector(bounds.W);
}

// --------------------------------------------
// Coordinate change from the new base to the old
inline void InvCoordinate_Change(const MatrixType &newBaseInv, const VectorType &newBaseOrigin, const VectorType &Z, VectorType &W)
{
	W = newBaseInv * Z + newBaseOrigin;
}

// --------------------------------------------
// Coordinate change fromt the old base to the new
inline void Coordinate_Change(const MatrixType &newBase, const VectorType &newBaseOrigin, const VectorType &W, VectorType &Z)
{
	Z = newBase * (W - newBaseOrigin);
}

// --------------------------------------------
// generating the tail
bool Generate_Tail(BoundsType &bounds, inclusion_Term &vField)
{
	Indent_PlusLength();

	bool go_on, first_run;
	TailType t_N, t_B;



	Indent_AddToOutput();
	ostream << "Generating topologically self-consistent bounds..." << std::endl;

	// checking isolation for tail nodes
	go_on = true;
	first_run = true;

	while ( go_on ) {
		go_on = false;

		for (int i = m + 1; i <= M; ++i) {

			// isolation check: Lambda * x_i^+ + N(x)_i < 0
			if (not( vField.Main_N(bounds.W, bounds.T, i) + vField.Eigenvalue(i) * bounds(i).rightBound() < 0)) {

				#ifdef FERI_DEBUG
					Indent_AddToOutput();
					ostream << "Node " << i << " is not consistent from the right. Check gives: " << vField.Main_N(bounds.W, bounds.T, i) + vField.Eigenvalue(i) * bounds(i).rightBound() << std::endl;
				#endif

				// modifying the tail node
				setRBound( bounds.T.mid(i) , ( right_infinc( - vField.Main_N(bounds.W, bounds.T, i) / vField.Eigenvalue(i) ) ).rightBound() );
				go_on = true;

			}

			// isolation check: Lambda * x_i^- + N(x)_i > 0
			if (not( vField.Main_N(bounds.W, bounds.T, i) + vField.Eigenvalue(i) * bounds(i).leftBound() > 0)) {

				#ifdef FERI_DEBUG
					Indent_AddToOutput();
					ostream << "Node " << i << " is not consistent from the left. Check gives: " << vField.Main_N(bounds.W, bounds.T, i) + vField.Eigenvalue(i) * bounds(i).leftBound() << std::endl;
				#endif

				// modifying the tail node
				setLBound( bounds.T.mid(i) , ( left_infinc( - vField.Main_N(bounds.W, bounds.T, i) / vField.Eigenvalue(i) ) ).leftBound() );
				go_on = true;
			}

		}



		// checking far tail
		vField.Tail_N(bounds.W, bounds.T, t_N);
		vField.Tail_b(bounds.W, bounds.T, t_B, t_N);

		if (not(bounds.T.contains_far_tail( t_B ) ) || first_run) {

			#ifdef FERI_DEBUG
				if ( not(first_run) ) {
					Indent_AddToOutput();
					ostream << "Fartail is not consistent." << std::endl;
				}

				Indent_AddToOutput();
				ostream << "Fartail of B:" << std::endl;
				t_B.print_farTail();

				Indent_AddToOutput();
				ostream << "Fartail of T:" << std::endl;
				bounds.T.print_farTail();
			#endif

			bounds.T.C = t_B.C * 1.1;
			bounds.T.s = t_B.s;
			go_on = true;
		}

		first_run = false;

	}

	return true;
	
	Indent_AddToOutput();
	ostream << "Topologically self-consistent bounds have been generated." << std::endl;

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "The starting bounds are now: " << std::endl;
		bounds.print();
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// transforming equation to new coordinates
void Calculate_Jacobian(MatrixType &Jacobian, MatrixType &JacobianErrors, const inclusion_Term &vField, const MatrixType &newBase, const MatrixType &newBaseInv, const VectorType &newBaseOrigin, const bVector &is_cpair_next)
{
	Indent_PlusLength();

	// The Jacobian
	Indent_AddToOutput();
	ostream << "Calculating Jacobian and separating errors..." << std::endl;
	Jacobian = newBase * vField[ newBaseOrigin ] * newBaseInv;

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "Rigorous Jacobian is: " << std::endl << Jacobian << std::endl;
	#endif

	// cutting Jacobian to the significant part and the perturbation: getting the block structure, and forming the complex boxes so the diagonal elements are the same, the off-diagonal elements are of the opposite sign
	for (int i = 1; i <= m; ++i)
		for (int j = 1; j <= m; ++j)

			if ( is_cpair_next(i) ) {

				// complex diagonal
				if (j == i) {

					JacobianErrors(i + 1, i + 1) = Jacobian(i + 1, i + 1) - Jacobian(i,i);
					Jacobian(i + 1, i + 1) = Jacobian(i,i);

				// complex block, non-diagonal
				} else if (j == i + 1) {

					JacobianErrors(j,i) = Jacobian(j,i) + Jacobian(i,j);
					Jacobian(j,i) = - Jacobian(i,j);

				// otherwise
				} else {

					JacobianErrors(i,j) = Jacobian(i,j);
					Jacobian(i,j) = 0;
				}

			} else if ((i > 1) && ( is_cpair_next(i - 1) )) {

				// otherwise
				if ( not((j == i) || (j == i - 1)) ) {
					JacobianErrors(i,j) = Jacobian(i,j);
					Jacobian(i,j) = 0;
				}

			// real non-diagonal
			} else if (i != j) {
				JacobianErrors(i,j) = Jacobian(i,j);
				Jacobian(i,j) = 0;
			}

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "The error matrix is: " << std::endl << JacobianErrors << std::endl;
		Indent_AddToOutput();
		ostream << "The Jacobian is: " << std::endl << Jacobian << std::endl;
	#endif

	Indent_AddToOutput();
	ostream << "The Jacobian is ready..." << std::endl;

	Indent_MinusLength();
}

// --------------------------------------------
// Calculate Galerkin errors
void Calculate_GalerkinErrors(const BoundsType &bounds, inclusion_Term &vField, const MatrixType &newBase, VectorType &GalerkinError)
{
	Indent_PlusLength();

	// Galerkin errors
	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "Calculating Galerkin Error...";
	#endif

	vField.EffectFromTail_in_PmF(bounds, GalerkinError);

	#ifdef FERI_DEBUG
		ostream << "done" << std::endl;
		Indent_AddToOutput();
		ostream << "Galerkin error: " << std::endl << GalerkinError << std::endl;
		Indent_AddToOutput();
		ostream << "Rotated Galerkin error: " << std::endl << newBase * GalerkinError << std::endl;
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// Calculate effect of higher order terms
void Calculate_EffectHigherOrder(const MatrixType &newBase, const MatrixType &newBaseInv, const VectorType &newBaseOrigin, inclusion_Term &vField, const BoundsType &bounds, VectorType &HighOrderTerms)
{
	Indent_PlusLength();

	MatrixType t_matrix;
	VectorType t_vec;



	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "Calculating higher order terms...";
	#endif

	for (int i = 1; i <= m; ++i) {
		t_matrix = vField.Hessian(i);
		t_vec = bounds.W - newBaseOrigin;
		HighOrderTerms(i) = interval(0);


		for (int j = 1; j <= m; ++j)
			for (int l = 1; l <= m; ++l)
				HighOrderTerms(i) += interval(0.5) * t_vec(j) * t_matrix(l,j) * t_vec(l);
	}

	#ifdef FERI_DEBUG
		ostream << "done" << std::endl;
		Indent_AddToOutput();
		ostream << "Higher order terms: " << std::endl << HighOrderTerms << std::endl;
		Indent_AddToOutput();
		ostream << "Rotated Higher order terms: " << std::endl << newBase * HighOrderTerms << std::endl;
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// Sum the constants
//  result = newBase * F(x_c) + JacobianErrors * y + HighOrder_innewbase + newBase * GalerkinError
void Sum_Constants(const MatrixType &newBase, const MatrixType &newBaseInv, const VectorType &newBaseOrigin, const inclusion_Term &vField, const MatrixType &JacobianErrors, const VectorType &Z, const VectorType &HighOrderTerms, const VectorType &GalerkinError, VectorType &SummedConstants)
{
	Indent_PlusLength();

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "Transforming Equation...";
	#endif

	SummedConstants = newBase * vField( newBaseOrigin ) + JacobianErrors * Z + newBase * HighOrderTerms + newBase * GalerkinError;

	#ifdef FERI_DEBUG
		ostream << "done" << std::endl;
		Indent_AddToOutput();
		ostream << "The Constant part is: " << std::endl << SummedConstants << std::endl;
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// Transform the Equation into new coordinates
void Transform_Equation(const MatrixType &newBase, const MatrixType &newBaseInv, const VectorType &newBaseOrigin, inclusion_Term &vField, const MatrixType &JacobianErrors, const BoundsType &bounds, const VectorType &Z, VectorType &HighOrderTerms, VectorType &GalerkinError, VectorType &SummedConstants)
{
	Indent_PlusLength();

	#ifdef FERI_DEBUG
		ostream << std::endl;
	#endif

	// calculating Galerkin errors
	Calculate_GalerkinErrors(bounds, vField, newBase, GalerkinError);

	// calculating the effect of higher order terms
	Calculate_EffectHigherOrder(newBase, newBaseInv, newBaseOrigin, vField, bounds, HighOrderTerms);

	// transforming the equation
	Sum_Constants(newBase, newBaseInv, newBaseOrigin, vField, JacobianErrors, Z, HighOrderTerms, GalerkinError, SummedConstants);

	#ifdef FERI_DEBUG
		ostream << std::endl;
	#endif

	Indent_MinusLength();
}

// --------------------------------------------
// Check for isolation in the main nodes
void Check_Isolation(const MatrixType &newBase, const MatrixType &newBaseInv, const VectorType &newBaseOrigin, const bVector &is_cpair_next, inclusion_Term &vField, const MatrixType &Jacobian, const MatrixType &JacobianErrors, BoundsType &bounds, VectorType &Z, VectorType &HighOrderTerms, VectorType &GalerkinError, VectorType &SummedConstants, VectorType &d, bVector &main_success)
{
	Indent_PlusLength();

	// max number of cycles to refine & try
	const int max_tries = 10;

	bool go_on;
	int iter;



	go_on = true;
	iter = 0;
	for (int i = 1; i <= m; ++i)
		main_success(i) = false;

	while (( go_on ) && (iter < max_tries)) {

		go_on = false;
		++iter;

		for (int i = 1; i <= m; ++i)

			// 2-dimensional eigenspace
			if ( is_cpair_next(i) ) {

				d(i) = (- sqrt( power(SummedConstants(i), 2) + power(SummedConstants(i + 1), 2) ) / Jacobian(i,i) ) * interval(-1,1);

				if ( Z(i).contains(d(i)) ) {

					if  ( Z(i) != d(i) ) {
						go_on = true;
						Z(i) = d(i);
						Z(i + 1) = d(i);
					}

					main_success(i) = true;
					main_success(i + 1) = true;

				} else {

					#ifdef FERI_DEBUG
						Indent_AddToOutput();
						ostream << "Isolation failed in the complex direction (" << i << "," << i + 1 << ") having " << d(i) << " not included in " << Z(i) << std::endl;
					#endif

					if ( main_success(i) ) {
						Indent_AddToOutput();
						ostream << "Serious error since isolation is lost at node " << i << " and " << i + 1 << std::endl;
						exit(0);
					}


					main_success(i) = false;
					main_success(i + 1) = false;

				}

				// 2nd coordinate is being skipped
				++i;

			// 1-dimensional eigenspace
			} else {

				d(i) = - SummedConstants(i) / Jacobian(i,i);

				if ( Z(i).contains( d(i) ) ) {

					if  ( Z(i) != d(i) ) {
						go_on = true;
						Z(i) = d(i);
					}

					main_success(i) = true;

				} else {

					#ifdef FERI_DEBUG
						Indent_AddToOutput();
						ostream << "Isolation failed in the real direction " << i << " having " << d(i) << " not included in " << Z(i) << std::endl;
					#endif

					if ( main_success(i) ) { 
						Indent_AddToOutput();
						ostream << "Serious error since isolation is lost at node " << i << " and " << i + 1 << std::endl;
						exit(0);
					}

					main_success(i) = false;

				}

			}

		#ifdef FERI_DEBUG
			Indent_AddToOutput();
			ostream << "Isolation status: " << std::endl << main_success << std::endl;
		#endif

	}

	Indent_MinusLength();
}

// --------------------------------------------
// Trying to prove the existence of a fixed point inside bounds
bool Validate_FixedPoint(BoundsType &bounds, const TailType &TheTail, inclusion_Term &vField, const double &nDELTA)
{

	// ------------------ INITIALIZATION

	Indent_PlusLength();

	// box size around candidate
	//const double nDELTA = 1e-7;
	VectorType Z;

	// rougher guess since around complex directions we use balls not boxes
	for (int i = 1; i <= m ; ++i)
		Z(i) = interval(-nDELTA, nDELTA);

	// max number of tries of blowing it
	const int max_tries = 5;

	// blowup factor
	const double blowup_factor = 2;



	Indent_AddToOutput();
	ostream << "Trying to validate fixed point..." << std::endl;
	Indent_AddToOutput();
	ostream << "Dimension of the projection is " << m << std::endl;
	Indent_AddToOutput();
	ostream << "Far tail starts from node " << M + 1 << std::endl;
	ostream << std::endl;
	Indent_AddToOutput();
	ostream << "Guessed fixed point is: " << std::endl;
	bounds.print();
	ostream << std::endl;
	Indent_AddToOutput();
	ostream << "Value of the vectorfield: " << std::endl << vField(bounds.W) << std::endl;
	ostream << std::endl;
	
// ------------------ END OF INITIALIZATION

// ------------------ BLOCK DECOMPOSITION

	// getting approximate Jacobian to dDPmFx0
	dMatrix dDPmFx0;

	Approximate_Jacobian(bounds, vField, dDPmFx0);



	// getting approximate eigenvalues and eigenvectors
	dMatrix dvectReal, dvectIm;
	dVector deigReal, deigIm;

	Eigenvalue_Decomposition(dDPmFx0, dvectReal, dvectIm, deigReal, deigIm);
	ostream << std::endl;



	// building up fundamental matrix with adding complex directions
	bVector is_cpair_next;

	Complex_Directions(is_cpair_next, dvectReal, dvectIm, deigReal, deigIm);
	ostream << std::endl;



	// defining new base
	MatrixType newBase, newBaseInv;
	VectorType newBaseOrigin;

	Define_Base(bounds, dvectReal, newBase, newBaseInv, newBaseOrigin);

	#ifdef FERI_DEBUG
		Indent_AddToOutput();
		ostream << "newBase J newBase^-1: " << std::endl << newBase * MatrixType(dDPmFx0) * newBaseInv << std::endl;
		Indent_AddToOutput();
		ostream << "newBase PmF(x_c): " << std::endl << newBase * vField(newBaseOrigin) << std::endl;
		ostream << std::endl;
	#endif

	// calculating and splitting the rigorous jacobian
	MatrixType Jacobian, JacobianErrors;

	Calculate_Jacobian(Jacobian, JacobianErrors, vField, newBase, newBaseInv, newBaseOrigin, is_cpair_next);
	ostream << std::endl;

	// Vectors for transforming the equation
	VectorType HighOrderTerms, GalerkinError, SummedConstants;
	
// ------------------ END OF BLOCK DECOMPOSITION

// ------------------ THE MAIN LOOP

	// obtaining isolation
	bool finished, tail_success;
	bVector main_success;
	VectorType d;
	int iter;

	finished = false;
	iter = 0;

	while (not(finished) && (iter < max_tries)) {

		finished = true;
		++iter;

	// ------------------ TOPOLOGICALLY SELF-CONSISTENT BOUNDS
	
		// we transform Z back to the original coordinates
		InvCoordinate_Change(newBaseInv, newBaseOrigin, Z, bounds.W);

		#ifdef FERI_DEBUG
			Indent_AddToOutput();
			ostream << "New main nodes in original coordinate system:" << std::endl;
			bounds.W.print();
			ostream << std::endl;
		#endif

		#ifdef FERI_GRAPHICS
	
			rootFrame.Clear();
			rootFrame << "The starting box";
			rootFrame.line(-40.0,0.,40.0,0.);
			rootFrame.line(0.,-40.0,0.,40.0);
			vField.sin_plot(bounds, 100);
			//waitBt();

		#endif

		// obtaining the tail
		bounds.T = TheTail;
		tail_success = Generate_Tail(bounds, vField);
		
		if (not(tail_success))
			finished = false;

	// ------------------ END OF TOPOLOGICALLY SELF-CONSISTENT BOUNDS

		if (tail_success) {

			#ifdef FERI_GRAPHICS
	
				rootFrame.Clear();
				rootFrame << "Tail is consistent";
				rootFrame.line(-40.0,0.,40.0,0.);
				rootFrame.line(0.,-40.0,0.,40.0);
				vField.sin_plot(bounds, 100);
				//waitBt();

			#endif

	// ------------------ TRANSFORMING INTO LOCAL COORDINATES
		
			// transforming the equation
			Transform_Equation(newBase, newBaseInv, newBaseOrigin, vField, JacobianErrors, bounds, Z, HighOrderTerms, GalerkinError, SummedConstants);

	// ------------------ END OF TRANSFORMING INTO LOCAL COORDINATES

	// ------------------ CHECK FOR ISOLATION

			// checking for isolation, refine if satisfied
			Check_Isolation(newBase, newBaseInv, newBaseOrigin, is_cpair_next, vField, Jacobian, JacobianErrors, bounds, Z, HighOrderTerms, GalerkinError, SummedConstants, d, main_success);
			ostream << std::endl;

			// checking result
			for (int i = 1; i <= m; ++i)
				if (not( main_success(i) ))
					finished = false;

	// ------------------ END OF CHECK FOR ISOLATION

	// ------------------ INCREASE IF FAILED

			if (not(finished)) {
				Indent_AddToOutput();
				ostream << "Verification failed for current main nodes at iteration: " << iter << std::endl;
				Indent_AddToOutput();
				ostream << "Isolation status is: " << main_success << std::endl;
				Indent_AddToOutput();
				ostream << "Blowing in non isolated directions..." << std::endl;
				ostream << std::endl;

				// increasing Z in the failed directions
				for (int i = 1; i <= m; ++i)
					if (not( main_success(i) )) {

						Z(i) = d(i) * blowup_factor;

						if (is_cpair_next(i)) {
							Z(i + 1) = Z(i);
							++i;
						}

					}

			}

	// ------------------ INCREASE IF FAILED

		}
		
	}
// ------------------ END OF THE MAIN LOOP

	if (finished) {

		// we transform Z back to the original coordinates
		InvCoordinate_Change(newBaseInv, newBaseOrigin, Z, bounds.W);

		Indent_AddToOutput();
		ostream << "Existence of fixed point is verified after " << iter << " iterations." << std::endl << std::endl;

		Indent_AddToOutput();
		ostream << "The first " << m << " coefficients of the approximate solution are: " << std::endl << "[";

		//Printing the center of the main part as a vector		

			int i;
			interval a, center;
			double radius;

			for(i = 1; i <= m; ++i) {
				a = (bounds.W)(i);
				split(a, center, radius);
				
				ostream << center.leftBound();
				if (i < m)
					ostream << ", ";
			}

			ostream << "]" << std::endl << std::endl;

		Indent_AddToOutput();
		ostream << "The fixed point is validated within the bounds: " << std::endl;
		bounds.print();

		// The various norms
		double c0normW, l2normW;
		
		bounds.calculateNorm(c0normW, l2normW);

		Indent_AddToOutput();
		ostream << "The maximal distance between the true solution and the candidate:" << std::endl;
		Indent_AddToOutput();
		ostream << "in C0 norm: " << c0normW << std::endl;
		Indent_AddToOutput();
		ostream << "in l2 norm: " << l2normW << std::endl;

	} else {

		Indent_AddToOutput();
		ostream << "--- Verification failed ---" << std::endl;

	}

	return (finished);

	Indent_MinusLength();
}

};

// ~ end of file ~
#endif
