#ifndef FERI_NR
#define FERI_NR
// --------------------------------------------
// -- Rigorous Numerics for Dissipative PDEs --
// --------------------------------------------
// ------------ Non-Rigorous Guess ------------
// --------------------------------------------
// by Ferenc Bartha
// http://www.math.u-szeged.hu/~barfer
// --------------------------------------------

#include "main.hpp"
#include "tools.hpp"
#include "structures.hpp"
#include "vectorfield.hpp"
#include "typedefs.hpp"

// --------------------------------------------

// non rigorous fixed point search with Newton method
void nonRig_newton(nRVectorField &my_nRField, const long int &maxstep, const double &stopNorm, const nRVector &xinit)
{
	Indent_PlusLength();

	nRVector xk, xk1, fvalue;
	nRMatrix jacobian, invjacobian;
	capd::vectalg::EuclNorm<nRVector, nRMatrix> nReNorm;
	long int newtonstep;



	newtonstep = 0;
	xk = xinit;
	xk1 = xk;
	fvalue = my_nRField(xk);

	Indent_AddToOutput();
	ostream << "Value of X: " << std::endl << xk << std::endl;
	Indent_AddToOutput();
	ostream << "Value of F: " << std::endl << fvalue << std::endl;
	Indent_AddToOutput();
	ostream << "Norm of F(X): " << nReNorm(fvalue) << std::endl;

	#ifdef FERI_GRAPHICS
		rootFrame.Clear();
		rootFrame.line(-40.0,0.,40.0,0.);
		rootFrame.line(0.,-40.0,0.,40.0);
		my_nRField.sin_plot(xk, 200);
		//waitBt();
	#endif

	while ((newtonstep < maxstep) && (nReNorm(fvalue) > stopNorm)) {
	
		jacobian = my_nRField[xk];
		invjacobian = capd::matrixAlgorithms::inverseMatrix(jacobian);
		fvalue =my_nRField(xk);
		xk1 = xk - invjacobian * fvalue;
		xk = xk1;

		Indent_AddToOutput();
		ostream << "Value of X: " << std::endl << xk << std::endl;
		Indent_AddToOutput();
		ostream << "Value of F: " << std::endl << fvalue << std::endl;
		Indent_AddToOutput();
		ostream << "Norm of F(X): " << nReNorm(fvalue) << std::endl;

		newtonstep++;

		#ifdef FERI_GRAPHICS
			rootFrame.Clear();
			rootFrame << "Step:" << newtonstep;
			rootFrame.line(-40.0,0.,40.0,0.);
			rootFrame.line(0.,-40.0,0.,40.0);
			my_nRField.sin_plot(xk, 200);
			//waitBt();
		#endif

	}

	if (nReNorm(fvalue) > stopNorm) {
		Indent_AddToOutput();
		ostream << "Zero not reached." << std::endl;
	} else {
		Indent_AddToOutput();
		ostream << "Zero reached." << std::endl;
	}

	Indent_MinusLength();
}

// non rigorous fixed point search with integration
void nonRig_integrate(nRVectorField &my_nRField, const long int &maxstep, const double &stopNorm, const nRVector &xinit)
{
	Indent_PlusLength();

	nRVector xk, fvalue;
	capd::vectalg::EuclNorm<nRVector, nRMatrix> nReNorm;
	long int step;
	double timestep;



	step = 0;
	my_nRField.initStep(timestep);

	nRSolver mySolver(my_nRField, GLOBAL_CONST_order, timestep);
	xk = xinit;
	fvalue = my_nRField(xk);

	Indent_AddToOutput();
	ostream << "Value of F: " << std::endl << fvalue << std::endl;
	Indent_AddToOutput();
	ostream << "Value of X: " << std::endl << xk << std::endl;
	Indent_AddToOutput();
	ostream << "Norm of F(X): " << nReNorm(fvalue) << std::endl;

	#ifdef FERI_GRAPHICS
		rootFrame.Clear();
		rootFrame.line(-40.0,0.,40.0,0.);
		rootFrame.line(0.,-40.0,0.,40.0);
		my_nRField.sin_plot(xk, 200);
		//waitBt();
	#endif

	while ((nReNorm(fvalue) > stopNorm)) {

		step++;

		xk = mySolver(xk);
		fvalue = my_nRField(xk);



		if (not(step % 10)) {

			#ifdef FERI_GRAPHICS
				rootFrame.Clear();
				rootFrame << "Step:" << step;
				rootFrame.line(-40.0,0.,40.0,0.);
				rootFrame.line(0.,-40.0,0.,40.0);
				my_nRField.sin_plot(xk, 200);
			#endif

			Indent_AddToOutput();
			ostream << "Value of F: " << std::endl << fvalue << std::endl;
			Indent_AddToOutput();
			ostream << "Value of X: " << std::endl << xk << std::endl;
			Indent_AddToOutput();
			ostream << "Norm of F(X): " << nReNorm(fvalue) << std::endl;
			
		}
	}

	#ifdef FERI_GRAPHICS
		rootFrame.Clear();
		rootFrame << "Integration is over (maxstep / zero is reached)";
		rootFrame.line(-40.0,0.,40.0,0.);
		rootFrame.line(0.,-40.0,0.,40.0);
		my_nRField.sin_plot(xk, 200);
		//waitBt();
	#endif
	
	Indent_AddToOutput();
	ostream << "Value of F: " << std::endl << fvalue << std::endl;
	Indent_AddToOutput();
	ostream << "Value of X: " << std::endl << xk << std::endl;
	Indent_AddToOutput();
	ostream << "Norm of F(X): " << nReNorm(fvalue) << std::endl;

	if (nReNorm(fvalue) > stopNorm) {
		Indent_AddToOutput();
		ostream << "Zero not reached." << std::endl;
	} else {
		Indent_AddToOutput();
		ostream << "Zero reached." << std::endl;
	}

	Indent_MinusLength();
}

// non rigorous fixed point
void nonRig_simulate(nRVectorField &my_nRField, const bool &isnewton, nRVector &xinit)
{
	const long int maxstep = 100000;
	const double stopNorm = 1e-14;

	if (isnewton) {
		nonRig_newton(my_nRField, maxstep, stopNorm, xinit);
	} else {
		nonRig_integrate(my_nRField, maxstep, stopNorm, xinit);
	}
}

// ~ end of file ~
#endif
