#ifndef FERI_STRUCTURES
#define FERI_STRUCTURES
// --------------------------------------------
// -- Rigorous Numerics for Dissipative PDEs --
// --------------------------------------------
// ---------------- Structures ----------------
// --------------------------------------------
// by Ferenc Bartha
// http://www.math.u-szeged.hu/~barfer
// --------------------------------------------

#include "main.hpp"
#include "tools.hpp"

// if defined, apriori enclosure is generated after each timestep
// #define FERI_AUTO_UPDATE_APRIORI

// if defined, the main nodes are being printed in standard form as well
#define FERI_PRINT_INTERVALFORM_ALSO

// --------------------------------------------

const int DEFAULT_TAIL_DECAY = 12;										// the default decay rate of a tail

// --------------------------------------------
// realizing the self consistent bounds and some operations on it

// the main nodes
template<int small_m>
class SelfC_Main : public capd::vectalg::Vector<interval, small_m>
{
	public:
		typedef capd::vectalg::Vector<interval, small_m>	BaseClass;
		typedef capd::vectalg::Vector<interval, small_m>	VectorType;

		// constructor
		SelfC_Main() : BaseClass() {}
		SelfC_Main(const int &dimension) : BaseClass(dimension) {}
		SelfC_Main(const VectorType &init) : BaseClass(init) {}
		SelfC_Main(const int &dimension, const bool &flag) : BaseClass(dimension, flag) {}

		// returns m
		int m()
		{
			return small_m;
		}

		// printing
		void print()
		{
			int i;
			double radius;
			interval a(0), center;



			#ifdef FERI_PRINT_INTERVALFORM_ALSO
				Indent_AddToOutput();
				ostream << "Main Part is defined by (in standard interval form):" << std::endl;
				for(i = 1; i <= small_m; ++i) {
					Indent_AddToOutput();
					ostream << "node " << i << ". = " << (*this)(i) << std::endl;
				}
				ostream << std::endl;
			#endif

			Indent_AddToOutput();
			ostream << "Main Part is defined by (in center-radius form):" << std::endl;
			for(i = 1; i <= small_m; ++i) {
				a = (*this)(i);
				split(a, center, radius);
				Indent_AddToOutput();
				ostream << "node " << i << ". = " << center.leftBound() << " + " << radius << " * [-1,1]" << std::endl;
			}
		}

		void print() const
		{
			int i;
			double radius;
			interval a(0), center;



			#ifdef FERI_PRINT_INTERVALFORM_ALSO
				Indent_AddToOutput();
				ostream << "Main Part is defined by (in standard interval form):" << std::endl;
				for(i = 1; i <= small_m; ++i) {
					Indent_AddToOutput();
					ostream << "node " << i << ". = " << (*this)(i) << std::endl;
				}
				ostream << std::endl;
			#endif

			Indent_AddToOutput();
			ostream << "Main Part is defined by (in center-radius form):" << std::endl;
			for(i = 1; i <= small_m; ++i) {
				a = (*this)(i);
				split(a, center, radius);
				Indent_AddToOutput();
				ostream << "node " << i << ". = " << center.leftBound() << " + " << radius << " * [-1,1]" << std::endl;
			}
		}


	protected:
};

// the mid tail
template<int small_m, int big_M>
class SelfC_midTail : public capd::vectalg::Vector<interval, big_M - small_m>
{
	public:
		typedef capd::vectalg::Vector<interval, big_M - small_m> BaseClass;

		using BaseClass::data;

		// returns a node
		interval& operator()(const int &i)
		{
			return (*this)[i - small_m - 1];
		}

		const interval& operator()(const int &i) const
		{
			return (*this)[i - small_m - 1];
		}

		// returns m
		int m()
		{
			return small_m;
		}

		// returns M
		int M()
		{
			return big_M;
		}

		// printing
		void print()
		{
			int i;



			Indent_AddToOutput();
			ostream << "Mid Tail is defined by:" << std::endl;
			for(i = small_m + 1; i <= big_M; ++i) {
				Indent_AddToOutput();
				ostream << "node " << i << ". = " << (*this)(i) << std::endl;
			}
		}

	protected:
};

// the tail
template<int small_m, int big_M>
class SelfC_Tail
{
	public:
		typedef SelfC_midTail<small_m, big_M> midTailType;

		interval s, C;
		midTailType mid;


		// constructor
		SelfC_Tail()
		{
			int i;



			s = interval(DEFAULT_TAIL_DECAY);
			C = interval(0);

			for (i = small_m + 1; i <= big_M; ++i)
				mid(i) = interval(0);
		}

		// decay rate modification for tail
		void decPower(const int &d)
		{
			s -= d;														// slower decay rate
			C = right( C / power_to_int(big_M + 1, d) );				// same bound at first far tail node
		}

		// returns a node, index is defined regarding the whole self consistent bounds structure
		interval operator()(const int &i)
		{
			if (i > big_M) {

				return ( C / power_to_int(i, s) * interval(-1,1) );			// node from the far tail

			} else if (i > small_m) {

				return mid(i);												// node from the mid tail

			} else {

				return interval(0);											// ERROR

			}
		}

		const interval operator()(const int &i) const
		{
			if (i > big_M) {

				return ( C / power_to_int(i, s) * interval(-1,1) );			// node from the far tail

			} else if (i > small_m) {

				return mid(i);												// node from the mid tail

			} else {

				return interval(0);											// ERROR

			}
		}

		// returns m
		int m()
		{
			return small_m;
		}

		// returns M
		int M()
		{
			return big_M;
		}

		const int M() const
		{
			return big_M;
		}

		// has mid tail
		bool has_midTail()
		{
			return (big_M > small_m);
		}

		// has empty far tail
		bool has_empty_farTail()
		{
			return (C == interval(0));
		}

		const bool has_empty_farTail() const
		{
			return (C == interval(0));
		}

		// contains the Tail B
		template<typename TailType>
		const bool contains(const TailType &B) const
		{
			int i;
			bool result;



			result = true;

			for (i = small_m + 1; i <= max( big_M, B.M() ) + 1; ++i) {
				result = result && (*this)(i).contains( B(i) );
			}

			if ( (*this).has_empty_farTail() && not( B.has_empty_farTail() ) ) {
				result = false;
			} else {
				result = result && (rightBound( (*this).s ) <= rightBound( B.s ));
			}

			return result;
		}

		template<typename TailType>
		const bool contains(const TailType &B, const int semmi) const
		{
			int i;
			bool result;



			result = true;

			for (i = small_m + 1; i <= max( big_M, B.M() ) + 1; ++i) {
				result = result && (*this)(i).contains( B(i) );
				
				if (not((*this)(i).contains( B(i) )))
					ostream << "NOT CONTAINED @ " << i << "    " << (*this)(i) << " NOTBIGGER THAN " << B(i) << std::endl;
			
			}

			if ( (*this).has_empty_farTail() && not( B.has_empty_farTail() ) ) {
				result = false;
			} else {
				result = result && (rightBound( (*this).s ) <= rightBound( B.s ));
			}

			return result;
		}

		// contains the far tail of B
		template<typename TailType>
		const bool contains_far_tail(const TailType &B) const
		{
			int i;
			bool result;



			result = (*this)( B.M() + 1 ).contains( B( B.M() + 1 ) );
			for (i = B.M() + 2; i <= big_M + 1; ++i)
				result = result && ( (*this)(i) ).contains( B(i) );

			if ( (*this).has_empty_farTail() && not( B.has_empty_farTail() ) ) {
				result = false;
			} else {
				result = result && (rightBound( (*this).s ) <= rightBound( B.s ));
			}

			return result;
		}

		// printing the far tail
		void print_farTail()
		{
			Indent_AddToOutput();
			ostream << "Far Tail is defined by:  C = " << C << " and s = " << s << std::endl;
			Indent_AddToOutput();
			ostream << "a[" << big_M + 1 << "] = " << (*this)(big_M + 1) << std::endl;
		}

		// printing
		void print()
		{
			mid.print();
			print_farTail();
		}

	protected:
};

// the self consistent bounds
template<int small_m, int big_M>
class SelfC_Bounds : public capd::diffIncl::InclRect2Set< capd::vectalg::Matrix<interval, small_m, small_m> >
{
	public:
		typedef capd::vectalg::Vector<interval, small_m>				VectorType;
		typedef capd::vectalg::Matrix<interval, small_m, small_m>		MatrixType;
		typedef capd::diffIncl::InclRect2Set< MatrixType >				BaseClass;
		typedef SelfC_Bounds<small_m, big_M>							BoundsType;
		typedef SelfC_Main<small_m>										MainType;
		typedef SelfC_Tail<small_m, big_M>								TailType;
		typedef SelfC_midTail<small_m, big_M>							midTailType;

		using BaseClass::get_x;
		using BaseClass::get_C;
		using BaseClass::get_r0;
		using BaseClass::get_B;
		using BaseClass::get_r;

		MainType W;
		TailType T;

		// constructors
		SelfC_Bounds() : BaseClass(small_m) {}
		SelfC_Bounds(int dimension) : BaseClass(dimension) {}
		SelfC_Bounds(const VectorType& the_x) : BaseClass(the_x) { W = VectorType(*this); }
		SelfC_Bounds(const VectorType& the_x, const VectorType& the_r0) : BaseClass(the_x, the_r0) { W = VectorType(*this); }
		SelfC_Bounds(const VectorType& the_x, const MatrixType& the_C, const VectorType& the_r0) : BaseClass(the_x, the_C, the_r0) { W = VectorType(*this); }
		SelfC_Bounds(const VectorType& the_x, const MatrixType& the_C, const VectorType& the_r0, const VectorType& the_r) : BaseClass(the_x, the_C, the_r0, the_r) { W = VectorType(*this); }
		SelfC_Bounds(const SelfC_Bounds& in_set) : BaseClass(in_set) {W = in_set.W; T = in_set.T;}

		// returns a node
		interval operator()(const int &i)
		{
			if (i > small_m) {
				return T(i);
			} else {
				return W(i);
			}
		}

		const interval operator()(const int &i) const
		{
			if (i > small_m) {
				return T(i);
			} else {
				return W(i);
			}
		}

		// returns m
		int m()
		{
			return small_m;
		}

		// returns M
		int M()
		{
			return big_M;
		}

		// printing
		void print()
		{
			W.print();
			T.print();
		}
		
		// getting the width of the enclosure and the norm of the solution
		void calculateNorm(double &c0normWidth, double &l2normWidth)
		{

			interval int_c0normWidth, int_l2normWidth, center;
			double radius;


			int_c0normWidth = 0;
			int_l2normWidth = 0;

		
			for(int i = 1; i <= big_M; ++i) {
				center = (*this)(i);
				split(center, center, radius);
				int_c0normWidth += radius;
				int_l2normWidth += power(interval(radius), 2);
			}
			c0normWidth = (int_c0normWidth + T.C / power_to_int(big_M, T.s - 1) / (T.s - 1)).rightBound(); 
			l2normWidth = sqrt((int_l2normWidth + power(T.C, 2) / power_to_int(big_M, 2 * T.s - 1) / (2 * T.s - 1)).right()).rightBound();
		}

		// moves the set, saves result separately
		template<typename DiffIncl>
		void move(DiffIncl &diffIncl, BoundsType &result) const
		{
			#ifdef FERI_DEBUG
				Indent_PlusLength();
				Indent_AddToOutput();
				ostream << "move: started" << std::endl;
			#endif

			// constructing a-priori enclosure
			diffIncl.constructNextStepAprioriDiffEnclosure(*this);

			#ifdef FERI_DEBUG
				Indent_AddToOutput();
				ostream << "move: CAPD diffIncl method has started";
			#endif

			// updating the tail
			diffIncl.moveTail((*this), result);

			// moving the main nodes
			BaseClass::move(diffIncl, result);

			#ifdef FERI_DEBUG
				ostream << std::endl;
				Indent_AddToOutput();
				ostream << "move: CAPD diffIncl method has finsihed" << std::endl;
			#endif

			result.W = MainType(result);

			#ifdef FERI_DEBUG
				Indent_AddToOutput();
				ostream << "move: finished" << std::endl;
				Indent_MinusLength();
			#endif
		}

		// moves the set, saves result separately, saves apriori
		template<typename DiffIncl>
		void move(DiffIncl &diffIncl, BoundsType &result, VectorType &bounds) const
		{
			move(diffIncl, result);
			bounds = diffIncl.mainEnclosure();
		}

		// moves the set
		template<typename DiffIncl>
		void move(DiffIncl &diffIncl)
		{
			BoundsType result(small_m);
			move(diffIncl, result);
			(*this) = result;
		}

		// moves the set, saves apriori
		template<typename DiffIncl>
		void move(DiffIncl &diffIncl, VectorType &bounds)
		{
			move(diffIncl);
			bounds = diffIncl.mainEnclosure();
		}

	protected:
	
};

// ~ end of file ~
#endif
