#ifndef FERI_VECTORFIELD
#define FERI_VECTORFIELD
// --------------------------------------------
// ------ Kuramoto-Shivashinsky Equation ------
// --------------------------------------------
// ------- Equation-Specific Routines ---------
// --------------------------------------------
// by Ferenc Bartha
// http://www.math.u-szeged.hu/~barfer
// --------------------------------------------
// Vectorfield for the destabilized KS

#include "main.hpp"
#include "tools.hpp"
#include "structures.hpp"

// --------------------------------------------
// the vectorfield
template<int small_m, int big_M>
class VectorField : public capd::dynsys::FadMap<interval,small_m>
{
	public:

		typedef typename capd::dynsys::FadMap<interval, small_m>::MatrixType MatrixType;
		typedef typename capd::dynsys::FadMap<interval, small_m>::VectorType VectorType;
		typedef typename SelfC_Bounds<small_m, big_M>::BoundsType BoundsType;
		typedef typename BoundsType::MainType MainType;
		typedef typename BoundsType::midTailType midTailType;
		typedef typename BoundsType::TailType TailType;



		int p;



		interval int_big_M;
		interval int_small_m;
		interval int_L;
		interval int_A;
		interval int_B;
		interval int_C;
		interval int_eta;
		interval int_gamma;

		
		
		// constructor
		VectorField (const interval &p_alpha, const interval &p_beta, const interval &p_gamma, const interval &p_nu, const interval &p_L)
		{
			int_big_M   = interval(big_M);
			int_small_m = interval(small_m);


			
			// [SPECIFIC]
			// Dissipativity rate
			p  = 4;


			
			// [SPECIFIC]
			// Parameters of the equation

			int_L = p_L;
			int_eta = 2 * capd::intervals::DoubleInterval::pi() / int_L;
			int_gamma = p_gamma;
			int_C = p_alpha;
			int_A = power(int_eta, 4) * p_nu;
			int_B = power(int_eta, 2) * p_beta;
		}

// --------------------------------------------

		// [SPECIFIC]
		// FadBad::vectorfield
		// u_t = P_m[ -(u_xxxx + \beta u_xx + \gamma uu_x) + \alpha u ]
		template <typename AVector>
		AVector operator()(const AVector &in) const
		{
			AVector out(small_m);
			int k, n;
			interval int_k, t;

			for(k = 1; k <= small_m; ++k) {
				int_k = interval(k);

				for(n = 1; n <= small_m - k; ++n) {
					out[k - 1] = out[k - 1] - in[n - 1] * in[n + k - 1];
				}

				for(n = 1; n <= k - 1; ++n) {
					out[k - 1] = out[k - 1] + (in[n - 1] * in[k - n - 1] / interval(2.0));
				}

				out[k - 1] = out[k - 1] * int_gamma * int_k * int_eta;

				out[k - 1] = out[k - 1] + (int_C + int_k * int_k * (int_B - int_A * int_k * int_k)) * in[k - 1];

			}

			return out;
		}

		// FadBad::derivative
		MatrixType operator[](const VectorType& u) const
		{
			MatrixType der(small_m, small_m);
			computeDerivative(*this, u, der);
			return der;
		}

		// FadBad::returns the dimension
		int getDimension() const
		{
			return small_m;
		}

// --------------------------------------------

		void sin_plot(const BoundsType &bounds, const int &Resolution)
		{
			int i;
			double x, L;
			interval point;


			L = int_L.leftBound();
			x = - L / 2.0;
			while(x <= L / 2.0) {

				point = bounds.T.C * interval(-1,1) / (power_to_int(big_M, bounds.T.s - 1) * (bounds.T.s - 1));

				for (i = big_M; i >= 1; --i)
					point += - 2.0 * bounds(i) * capd::intervals::sin( interval(2.0) * capd::intervals::DoubleInterval::pi() / interval(L) * interval(i) * interval(x));

				rootFrame.dot(x, point.rightBound(), RED);
				rootFrame.dot(x, point.leftBound(), BLUE);

				x += 1.0 / Resolution;

			}
		}

// --------------------------------------------

		// returns the k th node from a W + T form
		interval getNode(const MainType &W, const TailType &T, const int &k)
		{
			if (k <= small_m) {
				return W(k);
			} else {
				return T(k);
			}
		}

// --------------------------------------------

		// [SPECIFIC]
		// The eigenvalues of the linear part
		inline interval Eigenvalue(const int &k);				// returns the k th eigenvalue
		inline interval V(const int &k);						// computes V(k)

// --------------------------------------------
		
		// Galerkin-projection
		interval PmF(const MainType &W, const TailType &T, const int &k);								// computes PmF
		
		// Galerkin-error
		void EffectFromTail_in_PmF(const BoundsType &bounds, VectorType &result);						// computes effect from the tail in PmF
		
		// [SPECIFIC]
		// The k-th component of the Galerkin-error
		interval EffectFromTail_in_PmF(const MainType &W, const TailType &T, const int &k);				// computes effect from the tail in PmF

		// [SPECIFIC]
		// The Hessian for component k
		MatrixType Hessian(const int &k);
		
		// [SPECIFIC]
		// Nonlinear part
		interval Main_N(const MainType &W, const TailType &T, const int &k);							// computes N(k) for main node
		void Tail_N(const MainType &W, const TailType &T, TailType &N);									// computes N for tail nodes
		
		// b
		inline interval bi(const MainType &W, const TailType &T, const int &i);							// computes b(i)
		void Tail_b(const MainType &W, const TailType &T, TailType &b, TailType &N);					// computes b for tail nodes

// --------------------------------------------

	protected:

		TailType m_b, m_N, m_tail_center;

// --------------------------------------------

		// [SPECIFIC]
		// The formulas for the finite (FS) and infinite (IS) sum in N
		interval LowNode_FS(const MainType &W, const TailType &T, const int &k);						// computes FS(k) for k <= M
		interval HighNode_FS(const MainType &W, const TailType &T, const int &k);						// computes FS(k) for M < k <= 2M
		interval ExtraHighNode_Multiplied_FS(const MainType &W, const TailType &T, const int &k);		// computes FS(k) * k ^ (s - 1) for 2M < k
		interval ExtraHighNode_FS(const MainType &W, const TailType &T, const int &k);					// computes FS(k) for 2M < k

		interval LowNode_IS(const MainType &W, const TailType &T, const int &k);						// computes IS(k) for k <= M
		interval HighNode_Multiplied_IS(const MainType &W, const TailType &T, const int &k);			// computes IS(k) * k ^ (s - 1) for M < k
		interval HighNode_IS(const MainType &W, const TailType &T, const int &k);						// computes IS(k) for M < k

		// --------------------------------------------
		
		// [SPECIFIC]
		// computes effect from the tail in IS(k) for k <= M
		interval EffectFromTail_in_LowNode_IS(const MainType &W, const TailType &T, const int &k);		
};

// --------------------------------------------

// [SPECIFIC]
// returns the k th eigenvalue
template<int small_m, int big_M>
inline interval VectorField<small_m, big_M>::Eigenvalue(const int &k)
{
	interval int_k(k);



	return int_C + power_up(int_k, 2) * (int_B - int_A * power_up(int_k, 2));
}

// [SPECIFIC]
// computes V(k)
// \rho_k = - k^4 [ \eta^4 - \eta^2 \beta / k^2 - \alpha / k^4 ] = - k^4 [ int_A - int_B / k^2 - int_C / k^4 ]
template<int small_m, int big_M>
inline interval VectorField<small_m, big_M>::V(const int &k)
{
	return - int_C / power_to_int(k, 4) - int_B / power_to_int(k, 2) + int_A;
}

// --------------------------------------------

// computes PmF
// (k) = PmN(W + T)[k] + Lambda_k * a[k]
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::PmF(const MainType &W, const TailType &T, const int &k)
{
	return Main_N(W, T, k) + Eigenvalue(k) * W(k);
}

// computes effect from the tail in PmF - Galerkin error (that comes only from the nonlinear part, in our case from IS(k))
template<int small_m, int big_M>
void VectorField<small_m, big_M>::EffectFromTail_in_PmF(const BoundsType &bounds, VectorType &result)
{
	int k;



	for(k = 1; k <= small_m; ++k)
		result(k) = EffectFromTail_in_PmF(bounds.W, bounds.T, k);
}

// [SPECIFIC]
// computes effect from the tail in PmF - Galerkin error (that comes only from the nonlinear part, in our case from IS(k))
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::EffectFromTail_in_PmF(const MainType &W, const TailType &T, const int &k)
{
	interval int_k(k);



	return int_k * int_eta * int_gamma * EffectFromTail_in_LowNode_IS(W, T, k) ;
}

// [SPECIFIC]
// computes Hessian of the k-th component
template<int small_m, int big_M>
typename VectorField<small_m, big_M>::MatrixType VectorField<small_m, big_M>::Hessian(const int &k)
{
	int n;
	interval int_k;
	MatrixType H;


	int_k = interval(k);

	for(n = 1; n <= k - 1; ++n) {
			H(n, k - n) += 0.5;
			H(k - n, n) += 0.5;
	}
	for(n = 1; n <= small_m - k; ++n) {
			H(n, k + n) -= 1.0;
			H(k + n, n) -= 1.0;
	}
	
	H = H * int_eta * int_gamma * int_k;


	return H;
}

// [SPECIFIC]
// computes N(k) for a main node
// (k) = k * \eta * ( FS(k) / 2 - IS(k) )
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::Main_N(const MainType &W, const TailType &T, const int &k)
{
	interval int_k(k);



	return int_k * int_eta * int_gamma * ( LowNode_FS(W, T, k) / interval(2.0) - LowNode_IS(W, T, k) );
}

// [SPECIFIC]
// computes N for the tail
// D1_<=_2M = max_k { k^(s - 1) * FS(k) }
template<int small_m, int big_M>
void VectorField<small_m, big_M>::Tail_N(const MainType &W, const TailType &T, TailType &N)
{
	int k;
	interval D1_larger_2M, D1_smaller_2M, D1, D2, int_k;
	double L;



	// there is a mid-tail
	if (small_m < big_M) {
		for(k = small_m + 1; k <= big_M; ++k) {
			int_k = interval(k);
			N.mid(k) = int_k * int_eta * int_gamma * ( LowNode_FS(W, T, k) / interval(2.0) - LowNode_IS(W, T, k));
		}
	}

	
	
	// N for the far-tail

	// 2M >= k > M
	// An upper bound for any D from D / k^{s-1} for FS(k)
	D1_smaller_2M = right( power_to_int(int_big_M + 1, T.s - 1) * capd::abs( HighNode_FS(W, T, big_M + 1) ) );
	for(k = big_M + 2; k <= 2 * big_M; ++k) {
		int_k = interval(k);
		L = capd::max( rightBound(D1_smaller_2M), rightBound(power_to_int(int_k, T.s - 1) * capd::abs( HighNode_FS(W, T, k) )) );
		D1_smaller_2M = interval(L);
	}
	D1_smaller_2M = D1_smaller_2M;

	// k > 2M
	// D from D / k^{s-1} for FS(k)
	D1_larger_2M = right( ExtraHighNode_Multiplied_FS(W, T, k) );

	// Pick the bigger D for FS to have uniform bound
	L = capd::max( rightBound(D1_smaller_2M), rightBound(D1_larger_2M) );
	D1 = interval(L);

	// D from D / k^{s-1} for IS(k)
	D2 = right( HighNode_Multiplied_IS(W, T, k) );

	// N is given as
	N.s = T.s - 2;
	N.C = capd::abs(int_gamma * int_eta * right( D1 / interval(2.0) + D2 ));
}

// computes b(i) for main nodes
// N / - \rho
template<int small_m, int big_M>
inline interval VectorField<small_m, big_M>::bi(const MainType &W, const TailType &T, const int &i)
{
	return - Main_N(W, T, i) / Eigenvalue(i);
}

// computes b for all tail nodes
// N / - \rho
template<int small_m, int big_M>
void VectorField<small_m, big_M>::Tail_b(const MainType &W, const TailType &T, TailType &b, TailType &N)
{
	int i;



	// compute N
	Tail_N(W, T, N);

	// compute b
	for (i = small_m + 1; i <= big_M; ++i)
		b.mid(i) = - N(i) / Eigenvalue(i);

	b.s = N.s + p;
	b.C = right( N.C / V(big_M) );
}

// --------------------------------------------

// [SPECIFIC]
// computes FS(k) for k <= M
// \sum_{n = 1}^{k - 1} ( a[n] * a[k - n] )
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::LowNode_FS(const MainType &W, const TailType &T, const int &k)
{
	int n;
	interval result(0);

	for(n = 1; 2 * n < k; ++n)
		result += getNode(W, T, n) * getNode(W, T, k - n);

	result *= 2;

	if (not(k % 2))
		result += power_to_int( capd::abs(getNode(W, T, k / 2)), 2 );

	return result;
}

// [SPECIFIC]
// computes FS(k) for M < k <= 2M
// (( \sum_{n = 1}^{k - M - 1} ( |a[n]| / (k - n)^s * [-1,1] ) ) * C + \sum_{n = k - M}^{n < k / 2} ( a[n] * a[k - n] ) ) * 2 + e(k) * a[k / 2]^2
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::HighNode_FS(const MainType &W, const TailType &T, const int &k)
{
	int n;
	interval int_n, int_k(k), result(0), symmint(-1,1);



	for(n = 1; n <= k - big_M - 1; ++n) {
		int_n = interval(n);
		result += capd::abs( getNode(W, T, n) ) / power_to_int(int_k - int_n, T.s) * symmint;
	}

	result *= T.C;

	for(n = k - big_M; 2 * n < k; ++n)
		result += getNode(W, T, n) * getNode(W, T, k - n);

	result *= 2;

	if (not(k % 2))
		result += power_to_int( capd::abs(getNode(W, T, k / 2)), 2 );

	return result;
}

// [SPECIFIC]
// computes FS(k) * k ^ (s - 1) for 2M < k
// [ [ 1 / ( M^s * (s - 1) ) + ( [ 2 / (2M + 1) ]^s ) / (2M + 1) ] * C + [ \sum_{n = 1}^{M} ( |a[n]| ) ] * 2 / (2M + 1) ] * 2^s * C
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::ExtraHighNode_Multiplied_FS(const MainType &W, const TailType &T, const int &k)
{
	interval result, temp(0);
	int n;



	result = 1 / ( power_to_int(int_big_M, T.s) * (T.s - 1) );
	result += power_to_int(2 / (2 * int_big_M + 1), T.s) / (2 * int_big_M + 1);
	result *= T.C;

	for(n = 1; n <= big_M; ++n)
		temp += capd::abs( getNode(W, T, n) );
	temp *= 2;
	temp /= (2 * int_big_M + 1);

	result += temp;
	result *= power_to_int(2, T.s) * T.C;

	return result;
}

// [SPECIFIC]
// computes FS(k) for 2M < k
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::ExtraHighNode_FS(const MainType &W, const TailType &T, const int &k)
{
	return ExtraHighNode_Multiplied_FS(W, T, k) / power_to_int(k, T.s - 1);
}

// [SPECIFIC]
// computes IS(k) for k <= M
// [ C / [ (k + M + 1)^s * (s - 1) * M^(s - 1) ] * [-1,1] + \sum_{n = M - k + 1}^{M} ( |a[n]| / (k + n)^s * [-1,1] ) ] * C + \sum_{n = 1}^{M - k} ( a[n] * a[k + n] )
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::LowNode_IS(const MainType &W, const TailType &T, const int &k)
{
	interval int_k(k), symmint(-1,1), int_n, result;
	int n;


	result = T.C / (power_to_int(int_k + int_big_M + 1, T.s) * (T.s - 1) * power_to_int(int_big_M, T.s - 1));

	// positivity was guaranteed, but let's make sure
	result = capd::abs(result);
	
	for(n = big_M - k + 1; n <= big_M; ++n) {
		int_n = interval(n);
		result += capd::abs( getNode(W, T, n) ) / power_to_int(int_k + int_n, T.s);
	}

	result *= symmint * T.C;

	for(n = 1; n <= big_M - k; ++n)
		result += getNode(W, T, n) * getNode(W, T, k + n);

	return result;
}

// [SPECIFIC]
// computes IS(k) * k ^ (s - 1) for M < k
// C / (M + 1) * [ C / ( (M + 1)^(s - 1) * (s - 1) ) + \sum_{n = 1}^{M} ( |a[n]| ) ] * [-1,1]
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::HighNode_Multiplied_IS(const MainType &W, const TailType &T, const int &k)
{
	interval result, symmint(-1,1);
	int n;



	result = T.C / ( power_to_int(int_big_M + 1, T.s - 1) * (T.s - 1) );
	for(n = 1; n <= big_M; ++n)
		result += capd::abs( getNode(W, T, n) );
	result = T.C / (int_big_M + 1) * result * symmint;

	return result;
}

// [SPECIFIC]
// computes IS(k) for M < k
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::HighNode_IS(const MainType &W, const TailType &T, const int &k)
{
	return HighNode_Multiplied_IS(W, T, k) / power_to_int(k, T.s - 1);
}

// --------------------------------------------

// [SPECIFIC]
// computes effect from the tail in IS(k) for k <= m
// \sum_{n = m - k + 1}^\infty a_n a_{k + n}
// that is essentially
// IS(k) - \sum_{n = 1}^{m - k} a_n a_{k + n}
// that implies modifying the formula for IS(k) in accordance
template<int small_m, int big_M>
interval VectorField<small_m, big_M>::EffectFromTail_in_LowNode_IS(const MainType &W, const TailType &T, const int &k)
{
	interval int_k(k), symmint(-1,1), int_n, result;
	int n;

	
	
	// it is enough to cut from the finite sum part as m <= M, IS(k) follows	
	result = T.C / (power_to_int(int_k + int_big_M + 1, T.s) * (T.s - 1) * power_to_int(int_big_M, T.s - 1));

	// positivity was guaranteed, but let's make sure
	result = capd::abs(result);
	
	for(n = big_M - k + 1; n <= big_M; ++n) {
		int_n = interval(n);
		result += capd::abs( getNode(W, T, n) ) / power_to_int(int_k + int_n, T.s);
	}

	result *= symmint * T.C;

	// this is 0 if m = M
	for(n = small_m - k + 1; n <= big_M - k; ++n)
		result += getNode(W, T, n) * getNode(W, T, k + n);

	return result;
}

// --------------------------------------------
// Truncated VectorField
template<class Scalar, int D>
class nonRigVectorField : public capd::dynsys::FadMap<Scalar,D>
{
  public:
    typedef typename capd::dynsys::FadMap<Scalar,D>::MatrixType MatrixType;
    typedef typename capd::dynsys::FadMap<Scalar,D>::VectorType VectorType;

	Scalar nu;
	Scalar beta;
	Scalar alpha;
	Scalar gamma;
	Scalar L;
	Scalar PI;
	Scalar ETA;
	Scalar A;
	Scalar B;
	Scalar C;

	// constructor
	nonRigVectorField(Scalar p_alpha, Scalar p_beta, Scalar p_gamma, Scalar p_nu, Scalar p_L) {
		nu = p_nu;
		beta = p_beta;
		alpha = p_alpha;
		gamma = p_gamma;
		L = p_L;
		
		PI = 3.141592653589793;
		
		ETA = 2.0 * PI / L;
		A = nu * pow(ETA, 4);
		B = beta * pow(ETA, 2);
		C = alpha;
	}

	// [SPECIFIC]
	// The non-rigorous Galerkin-projection
	template <typename AVector>
	AVector operator()(const AVector &in) const
	{
		AVector out(D);
		int k, n;
		
		for(k = 1; k <= D; ++k) {
			
			// sum_1 ^ DIM-k        a_n a_n+k
			for(n = 1; n <= D - k; ++n) {
				out[k - 1] = out[k - 1] - in[n - 1] * in[n + k - 1];
			}		

			// 1/2  sum_1 ^ k          a_n a_k-n
			for(n = 1; n <= k - 1; ++n) {
				out[k - 1] = out[k - 1] + (in[n - 1] * in[k - n - 1] / 2.0);
			}

			out[k - 1] = out[k - 1] * gamma * k * ETA;
		
			out[k - 1] = out[k - 1] + (C + k * k * B - A * k * k * k * k) * in[k - 1];
		
		}

		return out;
	}

	// [SPECIFIC]
	// if we would integrate it
	void initStep(double &h)
	{
		h = 0.00001;
	}

	void sin_plot(const VectorType &in, const int &Resolution)
	{
		Scalar x, point, temp;
		
		x = - L / 2.0;

		while(x <= L / 2.0) {

			point = 0;
			for (int i = D; i >= 1; --i) {
				point +=  - 2 * in(i) * sin( ETA * i * x);
			}

			rootFrame.dot(x, point, RED);

			x += 1.0 / Resolution;

		}
	}
	
	MatrixType operator[](const VectorType& u) const
	{
		MatrixType der(D,D);
		computeDerivative(*this,u,der);
		return der;
	}

	int getDimension() const
	{
		return D;
	}
};

// ~ end of file ~
#endif
